import os
import sys
import pathlib
import subprocess
import shutil
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext


class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=""):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):
    def run(self):
        try:
            subprocess.check_output(["cmake", "--version"])
        except OSError:
            raise RuntimeError(
                "CMake must be installed to build the following extensions: "
                + ", ".join(e.name for e in self.extensions)
            )

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        extdir = os.path.dirname(self.get_ext_fullpath(ext.name))

        # required for auto-detection of auxiliary "native" libs
        if not extdir.endswith(os.path.sep):
            extdir += os.path.sep

        cmake_args = [
            "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=" + extdir,
            "-DPYTHON_EXECUTABLE=" + sys.executable,
        ]

        cfg = "Debug" if self.debug else "Release"
        build_args = ["--config", cfg]

        cmake_args += ["-DCMAKE_BUILD_TYPE=" + cfg, "-DBUILD_PYKSTARS=YES", "-GNinja"]
        build_args += ["--"]

        env = os.environ.copy()
        env["CXXFLAGS"] = '{} -DVERSION_INFO=\\"{}\\"'.format(
            env.get("CXXFLAGS", ""), self.distribution.get_version()
        )
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)

        if not os.path.exists(extdir):
            os.makedirs(extdir)

        subprocess.check_call(
            ["cmake", ext.sourcedir + "/../../"] + cmake_args,
            cwd=self.build_temp,
            env=env,
        )

        subprocess.check_call(
            ["cmake", "--build", ".", "--target", "pykstars"] + build_args,
            cwd=self.build_temp,
        )

        ext_file = subprocess.run(
            ["find", "-name", "pykstars*.so"],
            cwd=self.build_temp,
            check=True,
            capture_output=True,
        ).stdout.decode("utf-8")[:-1]

        shutil.copy2(os.path.join(self.build_temp, ext_file), extdir)


# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "readme.md").read_text()

setup(
    name="pykstars",
    version="0.0.2",
    author="Valentin Boettcher",
    author_email="hiro@protagon.space",
    description="Some python bindings for kstars. Primarily used to package DSO catalogs.",
    long_description_content_type="text/markdown",
    license="GPLv2+",
    long_description=README,
    include_package_data=True,
    ext_modules=[CMakeExtension("pykstars")],
    cmdclass=dict(build_ext=CMakeBuild),
    zip_safe=False,
)



#!/usr/bin/env python

import csv
import os
import matplotlib.pyplot as plt
import code
import scipy.signal
from numpy import *

def read_data_from_folder(data_folder):
    data = [] # initialize empty data structure

    pixel_scale = -1

    datasets = os.walk(data_folder).next()[1]

    learning_cutoff = 0 # time to ignore at the start to allow for learning

    for ds, d in enumerate(datasets):
        files = os.walk('./data/' + d).next()[2]
        print '===== DATASET: {} ====='.format(d)
        r = -1

        for f in files:
            csv_file = data_folder + '/' + d + '/' + f
            with open(csv_file, 'rb') as csvfile:

                data.append([])
                guider_started = False
                settling_started = False
                csvreader = csv.reader(csvfile, delimiter=',', quotechar='"')
                for row in csvreader:
                    if (len(row) == 1):
                        if (row[0].split(' ')[0] == 'Guiding'):
                            if (row[0].split(' ')[1] == 'Begins'):
                                r = r + 1
                                data[ds].append([])
                                data[ds][r] = dict()
                                data[ds][r]['Time'] = []
                                data[ds][r]['RARawDistance'] = []
                                data[ds][r]['DECRawDistance'] = []
                                data[ds][r]['RAGuideDistance'] = []
                                data[ds][r]['DECGuideDistance'] = []
                                data[ds][r]['SNR'] = []
                                guider_started = True

                            if (row[0].split(' ')[1] == 'Ends' and not r == -1):
                                time = array(data[ds][r]['Time'])
                                ra_dist = array(data[ds][r]['RARawDistance']) * data[ds][r]['PixelScale']
                                dec_dist = array(data[ds][r]['DECRawDistance']) * data[ds][r]['PixelScale']
                                ra_guide = array(data[ds][r]['RAGuideDistance']) * data[ds][r]['PixelScale']
                                dec_guide = array(data[ds][r]['DECGuideDistance']) * data[ds][r]['PixelScale']
                                snr = array(data[ds][r]['SNR'])

                                ra_acc = (ra_guide.cumsum() + ra_dist)
                                total_dist = sqrt(power(ra_dist,2) + power(dec_dist,2))

                                ra_rms = sqrt(mean(power(ra_dist[(time > learning_cutoff)],2)))
                                dec_rms = sqrt(mean(power(dec_dist[(time > learning_cutoff)],2)))
                                total_rms = sqrt(mean(power(total_dist[(time > learning_cutoff)],2)))

                                data[ds][r]['RARMS'] = ra_rms
                                data[ds][r]['DECRMS'] = dec_rms
                                data[ds][r]['TotalRMS'] = total_rms

                                if (len(time) > 0 and time[-1]/60 >= 10):
                                    print '{:<14} | RA RMS: {:1.3f} | {:<14} | Dec RMS: {:1.3f} | Total RMS: {:1.3f} | Duration:{: 4.0f} | RelPerf: {: 6.1f}'.format(data[ds][r]['RAAlgorithm'], ra_rms, data[ds][r]['DecAlgorithm'], dec_rms, total_rms, time[-1]/60, 100*(1-ra_rms/dec_rms))

                                guider_started = False

                    if (len(row) == 2):
                        if (row[1] == ' Settling started'):
                            settling_started = True
                        if (row[1] == ' Settling complete'):
                            settling_started = False

                    if (guider_started and 0 < len(row) < 18): # non-guiding lines
                        if (row[0][0:11] == 'Pixel scale'):
                            pixel_scale = float(row[0][14:18])
                            data[ds][r]['PixelScale'] = pixel_scale

                        if (row[0].split(' = ')[0] == 'X guide algorithm'):
                            data[ds][r]['RAAlgorithm'] = row[0].split(' = ')[1]

                        if (row[0].split(' = ')[0] == 'Y guide algorithm'):
                            data[ds][r]['DecAlgorithm'] = row[0].split(' = ')[1]

                    if (guider_started and not settling_started and len(row) == 18): # guiding lines
                        if(row[2] == 'Mount'): # not interested in DROPped frames
                            data[ds][r]['Time'].append(float(row[1]))
                            data[ds][r]['RARawDistance'].append(float(row[5]))
                            data[ds][r]['DECRawDistance'].append(float(row[6]))
                            data[ds][r]['RAGuideDistance'].append(float(row[7]))
                            data[ds][r]['DECGuideDistance'].append(float(row[8]))
                            data[ds][r]['SNR'].append(float(row[16]))
    return data

def sq_dist(a, b):
    return power((matrix(a).transpose() - matrix(b)),2)

def covSE(a, b, t, l):
    return t * exp(-sq_dist(a, b)/(2*l**2))

def cov(a, b):
    return covSE(a, b, 50, 100)

def main():
    data = read_data_from_folder('data')

    ds = 0
    r = 0

    time = array(data[ds][r]['Time'])
    ra_dist = array(data[ds][r]['RARawDistance']) * data[ds][r]['PixelScale']
    ra_guide = array(data[ds][r]['RAGuideDistance']) * data[ds][r]['PixelScale']
    ra_acc = (ra_guide.cumsum() + ra_dist)

    #x = matrix(linspace(0,2000,200))
    #X = matrix(time)
    #Y = matrix(ra_acc).T

    #m = cov(x,X) * linalg.solve(cov(X,X) + eye(X.shape[1]),Y)
    #v = cov(x,x) - cov(x,X) * linalg.solve(cov(X,X) + eye(X.shape[1]),cov(X,x))

    #code.interact(local=dict(globals(), **locals()))

    #plt.plot(time, ra_acc, 'r+')
    #plt.plot(x.T, m, '-b')
    #plt.plot(x.T, m+2*matrix(diag(v)).T, '-b')
    #plt.plot(x.T, m-2*matrix(diag(v)).T, '-b')

    for dataset in data:
        nr = 0
        maxtime = 0
        miny = 0
        maxy = 0
        minx = 0
        maxx = 0
        fig = False
        for run in dataset:
            if (len(run['Time']) > 0 and run['Time'][-1]/60 >= 10):
                nr = nr + 1
                if (maxtime < run['Time'][-1]):
                    maxtime = run['Time'][-1]
                ra_dist = array(run['RARawDistance']) * run['PixelScale']
                ra_guide = array(run['RAGuideDistance']) * run['PixelScale']
                #ra_dist = (ra_guide.cumsum() + ra_dist)
                time = array(run['Time'])
                if (miny > min(ra_dist)):
                    miny = min(ra_dist)
                if (maxy < max(ra_dist)):
                    maxy = max(ra_dist)
                if (minx > min(time)):
                    minx = min(time)
                if (maxx < max(time)):
                    maxx = max(time)
                fig = True
        if fig:
            plt.figure()
        r = 0
        for run in dataset:
            if (len(run['Time']) > 0 and run['Time'][-1]/60 >= 10):
                r = r + 1
                time = array(run['Time'])
                ra_dist = array(run['RARawDistance']) * run['PixelScale']
                ra_guide = array(run['RAGuideDistance']) * run['PixelScale']
                ra_acc = (ra_guide.cumsum() + ra_dist)

                h = scipy.signal.firwin(numtaps=10, cutoff=40, nyq=160)
                ra_filt = scipy.signal.lfilter(h, 1.0, ra_dist)

                ax = plt.subplot(nr, 1, r)
                plt.plot(time, ra_filt, 'r-')
                plt.plot(time, ra_dist, 'b+')
                #plt.plot(time, ra_acc, 'r-+')
                ax.set_ylim(miny, maxy)
                ax.set_xlim(minx, maxx)
                plt.title('{} | Dataset {} | Run {}'.format(run['RAAlgorithm'], data.index(dataset), dataset.index(run)))

    run = data[0][1]
    ##run = data[1][0]
    ##run = data[2][11]
    #run = data[3][7]

    time = array(run['Time'])
    ra_dist = array(run['RARawDistance'])
    dec_dist = array(run['DECRawDistance'])
    ra_guide = array(run['RAGuideDistance'])
    dec_guide = array(run['DECGuideDistance'])
    snr = array(run['SNR'])

    with open('dataset03.csv', 'wb') as csvfile:
        csvwriter = csv.writer(csvfile, delimiter=',', )
        csvfile.write('Time,RARawDistance,DECRawDistance,RAGuideDistance,DECGuideDistance,SNR\r\n')
        for i in range(len(time)):
            csvfile.write('{:0.3f},{:0.3f},{:0.3f},{:0.3f},{:0.3f},{:0.2f}\r\n'.format(time[i], ra_dist[i], dec_dist[i], ra_guide[i], dec_guide[i], snr[i]))

    plt.show()

if __name__ == "__main__":
    main()



#!/usr/bin/env python

from numpy import genfromtxt
import matplotlib.pyplot as plt

def read_data():

    spectrum_data = genfromtxt('spectrum_data.csv', delimiter=',') # read GP data from csv
    spectrum_data = spectrum_data[1:,:] # strip first line to remove header text

    periods = spectrum_data[:,0]
    amplitudes = spectrum_data[:,1]

    return periods, amplitudes

def update_plot(fig, axes, p1):

    periods, amplitudes = read_data()

    p1.set_data(periods, amplitudes)
    axes.relim()
    axes.autoscale_view(True,True,True)
    fig.canvas.draw()

def main():
    print("main function started")
    plt.ion()
    fig = plt.figure()
    axes = fig.add_subplot(111)
    p1, = plt.plot([],[],'ob')

    while True:
        try:
            update_plot(fig, axes, p1)
        except:
            pass
        plt.pause(1.0)

if __name__ == "__main__":
    main()



"""
This file can help developers to optimize the parameters of the guiding algorithm.

GPOpt is used to do Bayesian optimization on the hyperparameters, the range for
optimization and the parameters to optimize are selected by setting the bounds
accordingly.

The function evaluation is done by calling a binary file GuidePerformanceEval
with the parameters as command line arguments. The return values is parsed from
stdin as a float.

The scoring algorithm tries to never degrade performance (uses the worst value
if it is negative) and averages the performance in all other cases. This way
we achieve good average performance while not degrading performance for anyone.
"""

import subprocess
import math
import numpy as np
import os
from GPyOpt.methods import BayesianOptimization

def eval_func(input):
    parameters = np.exp(input[0])
    folder = os.path.abspath(os.path.join('..','..','contributions', 'MPI_IS_gaussian_process', 'tests', 'gaussian_process'))

    # prepares command line arguments for an evaluation
    def make_args(file):
        return ['./GuidePerformanceEval', os.path.join(folder, f)] + [str(param) for param in parameters]

    # calculates the score for a specific dataset file
    def file_score(file):
        return float(subprocess.check_output(make_args(file)))

    # find all files in the directory
    files = os.walk(folder).next()[2]

    # extract all dataset files for the performance testing
    valid_files = [f for f in files if f.startswith('performance_') and f.endswith('.txt')]

    # calculate the scores for all files
    scores = [file_score(f) for f in valid_files]

    # only optimize the average if we pass regression testing
    if np.min(scores) < 0.0:
        score = np.min(scores) # make sure we pass regression tests
    else:
        score = np.mean(scores) # try to achieve good average performance

    return score

def optimize():
    """
    the parameters are:
        pontrol_gain
        min_periods_for_inference
        min_move
        SE0KLengthScale
        SE0KSignalVariance
        PKLengthScale
        PKPeriodLength
        PKSignalVariance
        SE1KLengthScale
        SE1KSignalVariance
        min_periods_for_period_estimation
        points_for_approximation
        prediction_gain
    """

    # the bounds are used to select which parameters to optimize,
    # and in which range to optimize. If the range is zero (e.g., (1.0,1.0))
    # the parameter isn't optimized at all.
    bounds = [(0.7,0.7), (2.0,2.0), (0.2,0.2), (700,700), (18,18), (10,10), (200,200), (20,20), (25,25), (10,10), (2,2), (100,100), (0.5,0.5)]
    bounds = np.log(bounds)

    # evaluate the current parameter set
    opt_params = [0.7, 2, 0.2, 700, 18, 10, 200, 20, 25, 10, 2, 100, 0.5]
    print eval_func([np.log(opt_params)])

    # the initial_design_numdata defines how many random values are tested.
    # apparently this is the most important value and should be as high as affordable.
    # num_cores and batch_size define how many evaluations run simultaneously.
    myBopt = BayesianOptimization(f=eval_func, bounds=bounds, num_cores=6, batch_size=6, maximize=True, initial_design_numdata=24)
    myBopt.run_optimization(verbosity=True, maximize=True)
    print -myBopt.fx_opt
    opt_params = np.exp(myBopt.x_opt)

    print "{:.2f} {:.2f} {:.2f} {:.2f} {:.2f} {:.2f} {:.2f} {:.2f} {:.2f} {:.2f} {:.2f} {:.2f} {:.2f}".format(opt_params[0], opt_params[1], opt_params[2], opt_params[3], opt_params[4], opt_params[5], opt_params[6], opt_params[7], opt_params[8], opt_params[9], opt_params[10], opt_params[11], opt_params[12])

if __name__ == "__main__":
    optimize()



#!/usr/bin/env python3

import sys
import json
import os

def log_message(message):
    """Logs a message to a file in the user's home directory."""
    log_file = os.path.expanduser("~/post_job_log.txt")
    with open(log_file, "a") as f:
        f.write(f"{message}\n")

def main():
    """
    Parses JSON from stdin and logs relevant information from a post-job event.
    """
    try:
        # Read the JSON data from standard input
        input_data = sys.stdin.read()
        if not input_data:
            log_message("No data received from stdin.")
            return

        data = json.loads(input_data)

        script_type = data.get("script_type", "UNKNOWN")
        log_message(f"Script Type: {script_type}")

        if script_type == "POST_JOB":
            # Log job details
            job_info = data.get("job")
            if job_info:
                target_name = job_info.get("target_name", "Unknown Target")
                log_message(f"Job for target '{target_name}' completed.")

            # Log sequence progress
            progress_info = data.get("progress")
            if progress_info:
                completed = progress_info.get("completed", 0)
                total = progress_info.get("total", 0)
                log_message(f"Sequence Progress: {completed}/{total} jobs complete.")
                if completed == total:
                    log_message("All sequence jobs are complete!")

    except json.JSONDecodeError:
        log_message(f"Failed to decode JSON: {input_data}")
    except Exception as e:
        log_message(f"An unexpected error occurred: {e}")

if __name__ == "__main__":
    main()



#!/usr/bin/env python

import matplotlib.pyplot as plt
from numpy import genfromtxt

def read_data():

    measurement_data = genfromtxt('tm_data.csv', delimiter=',') # read GP data from csv
    measurement_data = measurement_data[1:,:] # strip first line to remove header text

    location = measurement_data[:,0]
    output = measurement_data[:,1]

    gp_data = genfromtxt('gp_data.csv', delimiter=',')
    gp_data = gp_data[1:,:]

    x_range = gp_data[:,0]
    mean = gp_data[:,1]
    std = gp_data[:,2]

    return location, output

def update_plot(fig, axes, p1):
    #print("update_plot() called")

    location, output = read_data()
    #plt.clf()
    #plt.plot(location, output, 'r+')
    #plt.plot(x_range, mean, '-b')
    #plt.plot(x_range, mean+2*std, ':b')
    #plt.plot(x_range, mean-2*std, ':b')

    p1.set_data(location,output)

    axes.relim()
    axes.autoscale_view(True,True,True)
    fig.canvas.draw()
    #plt.draw()




def main():
    print("main function started")
    plt.ion()
    fig = plt.figure()
    axes = fig.add_subplot(111)
    p1, = plt.plot([],[],'r+')

    while True:
        try:
            update_plot(fig, axes, p1)
        except:
            pass
        plt.pause(1.0)

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
KStars Test Runner
==================
Discovers all CTest-registered Qt Test binaries, runs them (optionally in
parallel), saves per-test XML logs, and prints a colourised summary showing
which tests passed, which failed, and exactly where failures occurred.

Usage examples
--------------
  # Run every test
  python3 Tests/run_tests.py

  # Run only tests whose name contains "scheduler"
  python3 Tests/run_tests.py -k scheduler

  # Run only tests tagged with the 'stable' CTest label, 4 workers
  python3 Tests/run_tests.py -L stable -j 4

  # Exclude UI tests
  python3 Tests/run_tests.py --exclude-label ui

  # Specify a custom build dir and log output directory
  python3 Tests/run_tests.py --build-dir /tmp/kstars-build --log-dir /tmp/logs

  # Stream output live and also keep XML logs
  python3 Tests/run_tests.py -k focus --verbose --xml
"""

from __future__ import annotations

import argparse
import json
import os
import subprocess
import sys
import time
import xml.etree.ElementTree as ET
from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Optional, Tuple


# ---------------------------------------------------------------------------
# ANSI colour helpers (auto-disabled when stdout is not a tty)
# ---------------------------------------------------------------------------

def _supports_color() -> bool:
    return hasattr(sys.stdout, "isatty") and sys.stdout.isatty()


_USE_COLOR = _supports_color()


def _c(code: str, text: str) -> str:
    if not _USE_COLOR:
        return text
    return f"\033[{code}m{text}\033[0m"


def green(t: str) -> str:  return _c("32", t)
def red(t: str) -> str:    return _c("31", t)
def yellow(t: str) -> str: return _c("33", t)
def cyan(t: str) -> str:   return _c("36", t)
def bold(t: str) -> str:   return _c("1",  t)
def dim(t: str) -> str:    return _c("2",  t)


# ---------------------------------------------------------------------------
# Data structures
# ---------------------------------------------------------------------------

@dataclass
class Failure:
    function: str
    data_tag: str
    file: str
    line: str
    message: str


@dataclass
class TestResult:
    name: str
    command: List[str]
    labels: List[str]
    status: str = "pending"   # passed | failed | crashed | timeout | skipped
    elapsed: float = 0.0
    log_file: Optional[Path] = None
    failures: List[Failure] = field(default_factory=list)
    pass_count: int = 0
    fail_count: int = 0
    skip_count: int = 0
    returncode: int = 0


# ---------------------------------------------------------------------------
# CTest discovery
# ---------------------------------------------------------------------------

def discover_tests(build_tests_dir: Path) -> List[Dict]:
    """Return raw list of test dicts from `ctest --show-only=json-v1`."""
    result = subprocess.run(
        ["ctest", "--show-only=json-v1"],
        cwd=build_tests_dir,
        capture_output=True,
        text=True,
    )
    if result.returncode != 0 and not result.stdout.strip():
        print(red(f"ERROR: ctest discovery failed in {build_tests_dir}"))
        print(result.stderr)
        sys.exit(1)
    try:
        data = json.loads(result.stdout)
    except json.JSONDecodeError as exc:
        print(red(f"ERROR: Failed to parse ctest JSON output: {exc}"))
        sys.exit(1)
    return data.get("tests", [])


def filter_tests(
    raw_tests: List[Dict],
    name_filter: Optional[str],
    include_labels: List[str],
    exclude_labels: List[str],
) -> List[Dict]:
    """Apply name substring and label filters."""
    result = []
    for t in raw_tests:
        name = t.get("name", "")
        labels = []
        # Labels live inside the properties list
        for prop in t.get("properties", []):
            if prop.get("name") == "LABELS":
                labels = prop.get("value", [])
                break

        if name_filter and name_filter.lower() not in name.lower():
            continue
        if include_labels:
            if not any(lbl in labels for lbl in include_labels):
                continue
        if exclude_labels:
            if any(lbl in labels for lbl in exclude_labels):
                continue

        t["_labels"] = labels
        result.append(t)
    return result


# ---------------------------------------------------------------------------
# XML parsing (Qt Test output format)
# ---------------------------------------------------------------------------

def _parse_one_xml_doc(xml_text: str) -> Tuple[int, int, int, List[Failure]]:
    """Parse a single Qt Test XML document chunk."""
    passes = fails = skips = 0
    failures: List[Failure] = []

    try:
        root = ET.fromstring(xml_text)
    except ET.ParseError:
        return passes, fails, skips, failures

    for tf in root.iter("TestFunction"):
        func_name = tf.attrib.get("name", "")

        for incident in tf.iter("Incident"):
            itype = incident.attrib.get("type", "")
            data_tag = ""
            for dt in incident.iter("DataTag"):
                data_tag = dt.text or ""

            if itype == "pass":
                passes += 1
            elif itype == "skip":
                skips += 1
            elif itype in ("fail", "xfail", "bfail"):
                fails += 1
                file_attr = incident.attrib.get("file", "")
                line_attr = incident.attrib.get("line", "")
                msg = ""
                for desc in incident.iter("Description"):
                    msg = (desc.text or "").strip()
                failures.append(Failure(
                    function=func_name,
                    data_tag=data_tag,
                    file=file_attr,
                    line=line_attr,
                    message=msg[:200],
                ))

    return passes, fails, skips, failures


def parse_qttest_xml(xml_path: Path) -> Tuple[int, int, int, List[Failure]]:
    """
    Parse a Qt Test XML file and return (pass_count, fail_count, skip_count, failures).

    Some test binaries run multiple QTest classes back-to-back and emit one
    XML document per class concatenated in the same file.  We split on the
    '<?xml' declaration boundary and parse each chunk independently.
    """
    passes = fails = skips = 0
    all_failures: List[Failure] = []

    try:
        text = xml_path.read_text(encoding="utf-8", errors="replace")
    except OSError:
        return passes, fails, skips, all_failures

    if not text.strip():
        return passes, fails, skips, all_failures

    # Split into individual XML documents on every '<?xml' occurrence
    import re
    chunks = re.split(r'(?=<\?xml\b)', text)
    for chunk in chunks:
        chunk = chunk.strip()
        if not chunk:
            continue
        p, f, s, failures = _parse_one_xml_doc(chunk)
        passes += p
        fails  += f
        skips  += s
        all_failures.extend(failures)

    return passes, fails, skips, all_failures


# ---------------------------------------------------------------------------
# Running a single test
# ---------------------------------------------------------------------------

def run_one_test(
    raw: Dict,
    log_dir: Path,
    timeout: int,
    verbose: bool,
    save_xml: bool,
) -> TestResult:
    name = raw["name"]
    command: List[str] = raw.get("command", [])
    labels: List[str] = raw.get("_labels", [])

    if not command:
        result = TestResult(name=name, command=[], labels=labels, status="skipped")
        return result

    xml_log = log_dir / f"{name}.xml"
    plain_log = log_dir / f"{name}.log"

    # Run the binary with Qt Test XML output flag.
    # Qt Test writes the XML result to stdout; KDE/Qt startup diagnostics go to stderr.
    # We capture them separately so the XML file stays clean.
    cmd = command + ["-v2", "-xml"]

    t_start = time.monotonic()
    try:
        proc = subprocess.run(
            cmd,
            capture_output=not verbose,
            text=True,
            timeout=timeout,
            cwd=log_dir,   # safe working dir for any temp files the test creates
        )
        elapsed = time.monotonic() - t_start
        returncode = proc.returncode

        stdout_data = proc.stdout or "" if not verbose else ""
        stderr_data = proc.stderr or "" if not verbose else ""

    except subprocess.TimeoutExpired:
        elapsed = timeout
        returncode = -999
        stdout_data = ""
        stderr_data = f"[TIMEOUT after {timeout}s]"
    except FileNotFoundError:
        elapsed = 0.0
        returncode = -1
        stdout_data = ""
        stderr_data = f"[BINARY NOT FOUND: {command[0]}]"

    # stdout contains the Qt Test XML output; write it directly.
    # If for some reason XML appears mixed with preamble text, strip to first <?xml.
    xml_content = stdout_data
    if xml_content and not xml_content.lstrip().startswith("<?xml"):
        xml_start = xml_content.find("<?xml")
        xml_content = xml_content[xml_start:] if xml_start != -1 else ""

    xml_log.write_text(xml_content, encoding="utf-8", errors="replace")
    plain_log.write_text(stderr_data, encoding="utf-8", errors="replace")

    # Parse the XML for detailed results
    passes, fails, skips, failure_details = parse_qttest_xml(xml_log)

    # Determine overall status
    if returncode == -999:
        status = "timeout"
    elif returncode == -1:
        status = "crashed"
    elif returncode != 0 or fails > 0:
        status = "failed"
    else:
        status = "passed"

    tr = TestResult(
        name=name,
        command=command,
        labels=labels,
        status=status,
        elapsed=elapsed,
        log_file=xml_log,
        failures=failure_details,
        pass_count=passes,
        fail_count=fails,
        skip_count=skips,
        returncode=returncode,
    )
    return tr


# ---------------------------------------------------------------------------
# Summary printing
# ---------------------------------------------------------------------------

STATUS_ICON = {
    "passed":  green("✅ PASS"),
    "failed":  red("❌ FAIL"),
    "crashed": red("🔴 CRASH"),
    "timeout": yellow("⏱  TIME"),
    "skipped": dim("⏭  SKIP"),
    "pending": dim("…  PEND"),
}


def print_summary(results: List[TestResult], log_dir: Path, total_elapsed: float) -> None:
    sep = "─" * 72

    print()
    print(bold(sep))
    print(bold("  KStars Test Suite — Summary"))
    print(bold(sep))

    # Group by status for ordering: failed/crash/timeout first, then passed, skipped last
    ordered = (
        [r for r in results if r.status in ("failed", "crashed", "timeout")] +
        [r for r in results if r.status == "passed"] +
        [r for r in results if r.status == "skipped"]
    )

    for r in ordered:
        icon = STATUS_ICON.get(r.status, "?")
        elapsed_str = f"{r.elapsed:6.1f}s"
        counts = f"{green(str(r.pass_count)+'p')} {red(str(r.fail_count)+'f')} {dim(str(r.skip_count)+'s')}" \
                 if r.pass_count + r.fail_count + r.skip_count > 0 else ""
        label_str = dim(f"[{', '.join(r.labels)}]") if r.labels else ""
        print(f"  {icon}  {bold(r.name):<42} {elapsed_str}  {counts}  {label_str}")

        # Print failure details indented
        for f in r.failures:
            tag = f" [{f.data_tag}]" if f.data_tag else ""
            loc = f"{f.file}:{f.line}" if f.file else "unknown location"
            print(f"         {red('↳')} {cyan(f.function)}{yellow(tag)}")
            print(f"           {dim('at')} {loc}")
            if f.message:
                # Print first line of message
                first_line = f.message.split("\n")[0].strip()
                print(f"           {dim(first_line)}")

        if r.status in ("crashed", "timeout") and r.log_file:
            print(f"         {dim('log:')} {r.log_file}")

    print(bold(sep))

    # Aggregate counts
    total   = len(results)
    passed  = sum(1 for r in results if r.status == "passed")
    failed  = sum(1 for r in results if r.status == "failed")
    crashed = sum(1 for r in results if r.status == "crashed")
    timeout = sum(1 for r in results if r.status == "timeout")
    skipped = sum(1 for r in results if r.status == "skipped")

    parts = [
        green(f"{passed} passed"),
        red(f"{failed} failed") if failed else None,
        red(f"{crashed} crashed") if crashed else None,
        yellow(f"{timeout} timed out") if timeout else None,
        dim(f"{skipped} skipped") if skipped else None,
    ]
    summary_line = "  " + "  |  ".join(p for p in parts if p is not None)
    summary_line += f"  |  {dim(f'{total_elapsed:.1f}s total')}"
    print(summary_line)
    print(bold(sep))
    print(f"  {dim('Logs saved to:')} {log_dir}")
    print(bold(sep))
    print()


# ---------------------------------------------------------------------------
# Plain-text summary (written to summary.txt)
# ---------------------------------------------------------------------------

def write_plain_summary(results: List[TestResult], log_dir: Path, total_elapsed: float) -> None:
    lines: List[str] = []
    lines.append("KStars Test Suite — Summary")
    lines.append("=" * 72)

    for r in results:
        status = r.status.upper()
        lines.append(f"[{status:<7}] {r.name}  ({r.elapsed:.1f}s)  pass={r.pass_count} fail={r.fail_count} skip={r.skip_count}")
        for f in r.failures:
            tag = f" [{f.data_tag}]" if f.data_tag else ""
            loc = f"{f.file}:{f.line}" if f.file else "unknown location"
            lines.append(f"         FAILURE: {f.function}{tag}")
            lines.append(f"                  at {loc}")
            if f.message:
                lines.append(f"                  {f.message.split(chr(10))[0].strip()}")

    lines.append("=" * 72)
    passed  = sum(1 for r in results if r.status == "passed")
    failed  = sum(1 for r in results if r.status in ("failed", "crashed", "timeout"))
    skipped = sum(1 for r in results if r.status == "skipped")
    lines.append(f"Total: {len(results)}  Passed: {passed}  Failed: {failed}  Skipped: {skipped}  Time: {total_elapsed:.1f}s")

    (log_dir / "summary.txt").write_text("\n".join(lines) + "\n", encoding="utf-8")


# ---------------------------------------------------------------------------
# Build step
# ---------------------------------------------------------------------------

def build_tests(build_dir: Path, jobs: int) -> None:
    print(bold(f"Building tests in {build_dir} …"))
    result = subprocess.run(
        ["cmake", "--build", str(build_dir), "--target", "all", f"-j{jobs}"],
        cwd=build_dir,
    )
    if result.returncode != 0:
        print(red("Build failed — aborting."))
        sys.exit(result.returncode)
    print(green("Build succeeded.\n"))


# ---------------------------------------------------------------------------
# CLI
# ---------------------------------------------------------------------------

def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="KStars test runner — runs CTest Qt Test binaries, logs output, prints summary.",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=__doc__,
    )

    # Discovery / filtering
    parser.add_argument(
        "-k", "--filter", metavar="PATTERN",
        help="Only run tests whose name contains PATTERN (case-insensitive substring).",
    )
    parser.add_argument(
        "-L", "--label", metavar="LABEL", action="append", default=[],
        help="Only run tests that have this CTest label. May be repeated.",
    )
    parser.add_argument(
        "--exclude-label", metavar="LABEL", action="append", default=[],
        help="Exclude tests that have this CTest label. May be repeated.",
    )

    # Execution
    parser.add_argument(
        "-j", "--jobs", type=int, default=1, metavar="N",
        help="Run N tests in parallel (default: 1).",
    )
    parser.add_argument(
        "--timeout", type=int, default=120, metavar="SECONDS",
        help="Per-test timeout in seconds (default: 120).",
    )

    # Paths
    repo_root = Path(__file__).resolve().parent.parent
    default_build = repo_root / "build"
    parser.add_argument(
        "--build-dir", type=Path, default=default_build, metavar="DIR",
        help=f"CMake build directory (default: {default_build}).",
    )
    parser.add_argument(
        "--log-dir", type=Path, default=None, metavar="DIR",
        help="Directory for log files (default: ./test-logs-<timestamp>/).",
    )

    # Misc
    parser.add_argument(
        "--no-build", action="store_true",
        help="Skip the cmake --build step.",
    )
    parser.add_argument(
        "--xml", action="store_true",
        help="Preserve raw Qt XML output alongside .log files.",
    )
    parser.add_argument(
        "--verbose", action="store_true",
        help="Stream test output live to the terminal (implies --jobs 1).",
    )
    parser.add_argument(
        "--list", action="store_true",
        help="List discovered tests without running them, then exit.",
    )

    return parser.parse_args()


# ---------------------------------------------------------------------------
# Main
# ---------------------------------------------------------------------------

def main() -> None:
    args = parse_args()

    # Resolve build/Tests dir
    build_dir: Path = args.build_dir.resolve()
    build_tests_dir: Path = build_dir / "Tests"

    if not build_tests_dir.is_dir():
        print(red(f"ERROR: build/Tests directory not found: {build_tests_dir}"))
        print("Make sure you have configured with -DBUILD_TESTING=ON and built the project.")
        sys.exit(1)

    # Optionally build first
    if not args.no_build:
        build_tests(build_dir, args.jobs)

    # Discover tests
    print(bold("Discovering tests …"))
    raw_tests = discover_tests(build_tests_dir)
    filtered  = filter_tests(raw_tests, args.filter, args.label, args.exclude_label)

    if not filtered:
        print(yellow("No tests matched the given filters."))
        sys.exit(0)

    # List mode
    if args.list:
        print(bold(f"\nDiscovered {len(filtered)} test(s):"))
        for t in filtered:
            labels = t.get("_labels", [])
            label_str = f"  [{', '.join(labels)}]" if labels else ""
            print(f"  {t['name']}{label_str}")
        sys.exit(0)

    # Set up log dir
    if args.log_dir is None:
        stamp = datetime.now().strftime("%Y%m%d-%H%M%S")
        args.log_dir = Path.cwd() / f"test-logs-{stamp}"
    log_dir: Path = args.log_dir
    log_dir.mkdir(parents=True, exist_ok=True)

    # Verbose forces single-threaded (output would interleave otherwise)
    jobs = 1 if args.verbose else args.jobs

    print(bold(f"Running {len(filtered)} test(s)  [jobs={jobs}  timeout={args.timeout}s]"))
    print(dim(f"Logs → {log_dir}"))
    print()

    results: List[TestResult] = []
    global_start = time.monotonic()

    def _run(raw: Dict) -> TestResult:
        r = run_one_test(raw, log_dir, args.timeout, args.verbose, args.xml)
        icon = STATUS_ICON.get(r.status, "?")
        # Brief live status line
        if not args.verbose:
            elapsed_str = f"{r.elapsed:.1f}s"
            print(f"  {icon}  {r.name:<45} {elapsed_str}")
        return r

    if jobs == 1:
        for raw in filtered:
            results.append(_run(raw))
    else:
        futures = {}
        with ThreadPoolExecutor(max_workers=jobs) as ex:
            for raw in filtered:
                futures[ex.submit(_run, raw)] = raw["name"]
            for fut in as_completed(futures):
                results.append(fut.result())

    total_elapsed = time.monotonic() - global_start

    # Sort results back into discovery order for the summary
    order = {t["name"]: i for i, t in enumerate(filtered)}
    results.sort(key=lambda r: order.get(r.name, 9999))

    # Print and save summary
    print_summary(results, log_dir, total_elapsed)
    write_plain_summary(results, log_dir, total_elapsed)

    # Exit with non-zero if any test didn't pass
    any_bad = any(r.status not in ("passed", "skipped") for r in results)
    sys.exit(1 if any_bad else 0)


if __name__ == "__main__":
    main()
